\ recursive descent parser generator )

\ Copyright 1990, 1991, 1994, 2003 Martin Anton Ertl

\     This program is free software; you can redistribute it and/or modify
\     it under the terms of the GNU General Public License as published by
\     the Free Software Foundation; either version 2 of the License, or
\     (at your option) any later version.

\     This program is distributed in the hope that it will be useful,
\     but WITHOUT ANY WARRANTY; without even the implied warranty of
\     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
\     GNU General Public License for more details.

\     You should have received a copy of the GNU General Public License
\     along with this program; if not, write to the Free Software
\     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

\ ---[ Changes ]---------------------------------------------------------------

\ 29/5/08 Introduced new version of reduce from comp.lang.forth discussion

\ -----------------------------------------------------------------------------

.( Loading graycore.fth which includes code from ) cr 
.( Gray ... Copyright 1990-2003 Martin Anton Ertl; NO WARRANTY ) cr

\ ---[ Misc ]------------------------------------------------------------------

: no-syntax-expr-op drop ;    \ drop as Mini-oof methods are being used

: drop-args 2drop ;           \ 2drop as Mini-oof methods are being used

1 cells constant cell
s" address-unit-bits" environment? 0=
[IF]
   .( environmental attribute address-unit-bits unknown, assuming 8 bits/au ) cr
   .( if your machine has fewer bits/au, gray will not work ) cr
   \ if your machine has more bits/au, this assumption wastes space
   8
[THEN]
cells constant bits/cell

: 2, ( w1 w2 -- )
   here 2 cells allot 2!
;

: ?pairs ( n1 n2 -- )  \ aborts, if the numbers are not equal
   <> abort" mismatched parenthesis"
; 

: ',  ( -- ) ( use: ', name )
   ' ,
;

\ ---[ Stack administration ]--------------------------------------------------
\ this implementation is completely unsafe

\ use: n stack word
\ creates a stack called word with n cells
\ the first cell is the stackpointer

: stack  ( n -- )
   create here , cells allot
;

: push  ( n stack -- )  cell over +! @ ! ;

: top  ( stack -- n )  @ @ ;

: pop  ( stack -- )  [ -1 cells ] literal swap +! ;

: clear?  ( stack -- f )  dup @ = ;

: clear  ( stack -- )  dup ! ;

\ ---[ Sets ]------------------------------------------------------------------
\ Requires that sets.fth has been included

0 value tokset
0 value empty

\ max-member creates a new Set class called tokset, the size is unknown when
\ Gray is compiled. It must be called by the user before any sets are created
\ and any set operation. u is the highest member number of the set

: max-member  ( u -- )
   s" :set (ts) (ts) to tokset" evaluate
   tokset new to empty
;

\ Makes a set that contains only member u

: singleton  ( u -- set )
   tokset new tuck add-member
; 

\ Create new tokset set2 and copy contents of set1 into set2
 
: clone-set  ( set1 -- set2 )
   tokset new tuck copy-set      ( -- set2 )
;

\ ---[ Context management ]----------------------------------------------------

\ this stack holds the syntax-exprs currently being treated
\ enlarge it, if your grammar is large and complex

1000 stack context-stack
context-stack clear

\ get current syntax-expr )

: this  ( -- syntax-expr )  context-stack top ;

: new-context  ( syntax-expr -- )  context-stack push ;

: old-context  ( -- )  context-stack pop ;

\ ---[ Abstract generator classes ]----------------------------------------------
\ Subclasses will provide the concrete generate() methods for the individual
\ Syntax-Expr classes. These classes are declared here to provide the interface

object class
   class-method generate()
   class-method begin-def
   class-method end-def
end-class GeneratorBase

1 constant development

development [if]
:noname  ( class -- )
   abort" Method generate() not yet implemented"
; GeneratorBase defines generate()

:noname  ( class -- )
   abort" Method begin-def not yet implemented"
; GeneratorBase defines begin-def

:noname  ( class -- )
   abort" Method end-def not yet implemented"
; GeneratorBase defines end-def
[then]

GeneratorBase class
   class-method begin-action
   class-method end-action
end-class ActionGeneratorBase

development [if]
:noname  ( class -- )
   abort" Method begin-action not yet implemented"
; ActionGeneratorBase defines begin-action

:noname  ( class -- )
   abort" Method end-action not yet implemented"
; ActionGeneratorBase defines end-action
[then]

GeneratorBase class
   class-method code-nt
end-class Pass2GeneratorBase

development [if]
:noname  ( class -- )
   abort" Method code-nt not yet implemented"
; Pass2GeneratorBase defines code-nt
[then]

\ ---[ syntax-expr base class ]------------------------------------------------

object class
   cell var (mark-propagate) \ used to ensure that "propagate" is
                             \ called at least once for each syntax-expr
   cell var (mark-pass2)     \ make sure pass2 is called exactly once
   cell var (maybe-empty)    \ true if the syntax-expr can derive eps
   cell var (first-set)      \ all tokens a non-empty path may begin with, if 0
                             \ the first-set has not been computed yet
   cell var (follow-set)     \ terminal tokens that can follow the syntax-expr
   cell var (set-index)      \ first-set array index in generated Forth code
   cell class-var generator  \ Points to the concrete generator class
   method compute()
   method propagate()
   method pass2()
 \ !! replace the rest of the class with something working on your system
   2 cells var source-location   \ for error msgs
   cell var source->in
end-class Syntax-Expr

Pass2GeneratorBase Syntax-Expr generator !  \ Set to "not yet implemented"

\ get-generator obtains the generator class for a syntax expression. It uses
\ the fact that the first cell of an object points to its Mini-OOF class

: get-generator  ( syntax-expr -- gen-class )  @ generator @ ;

80 constant ebuf-size      \ To hold source code

\ getSource is introduced so that when the input format is BNF the details
\ of the BNF source text can be provided externally by re-ISing the deferred
\ word. The default gets the data from the input source (Gray format) 

defer getSource   ( -- caddr1 u1 u2 ) \ (caddr1 u1) is the source line
                                      \ u2 is the cursor index (>IN)

:noname  ( -- caddr1 u1 u2 )  source >in @ ; is getSource

\ set-source allows the user to define an external source e.g. for BNF input
\ If this is left in it needs a different name - clash with inputsources.fth
\ : set-source  ( xt1 -- )  ['] getSource defer! ;

\ Gray originally saved a line of text for each syntax-expr which led to a
\ line of text that creates several syntax-expr's saving that line of text
\ several times. saveSource now remembers the last line of text so that any
\ given line is saved once only. 

2variable last-textline here 0 last-textline 2!

: saveSource  ( syntax-expr -- )
   >r getSource                           ( -- caddr1 u1 u2 )
   r@ source->in ! ebuf-size min          ( -- caddr1 u1' )
   2dup last-textline 2@ compare
   if       \ Not yet saved, so save it
      here over last-textline 2!
      chars here over aligned allot       ( -- caddr1 u1" caddr2 )
      swap move                           ( -- )
   else
      2drop
   then
   last-textline 2@ r> source-location 2! ( -- )
;

:noname  ( class -- syntax-expr )
   [ object :: new ] >r                ( R: -- syntax-expr )
   false r@ 2dup (mark-propagate) !
   2dup (mark-pass2) !
   (maybe-empty) !
   0 r@ 2dup (first-set) !
   (set-index) !
   tokset new r@ (follow-set) !   \ Points to an empty set
   \ source location. !! replace the stuff until `;' with your stuff
   r@ saveSource r>                    ( -- syntax-expr )

\   getSource                           ( -- caddr1 u1 u2 )
\   r@ source->in ! ebuf-size min       ( -- caddr1 u1' )
\   here over r@ source-location 2!
\   chars here over aligned allot       ( -- caddr1 u1" caddr2 )
\   >r
\   cr ." Saving: " 2dup type cr
\   r>
\   swap move r>                        ( -- syntax-expr )

; Syntax-Expr defines new

\ ---[ Interface to syntax-expr data and methods ]-----------------------------
\ Done this way for easy evolution from the original Gray
\ May remove the context stack eventually whereupon these will disappear at
\ the cost of keeping syntax-expr's on the stack
\ They simply factor out the use of this

: mark-propagate    this (mark-propagate) ;
: mark-pass2        this (mark-pass2)     ;
: maybe-empty       this (maybe-empty)    ;
: first-set         this (first-set)      ;
: follow-set        this (follow-set)     ;
: set-index         this (set-index)      ;

\ ---[ Warnings and errors ]---------------------------------------------------

\ prints the info stored in source-location in a usable way
\ prints where the error happened
\ !! again, do it your way

: .in  ( -- )
   this source-location 2@ cr type cr
   this source->in @ 2 - spaces ." ^" cr  ." ::: "
;

: gray-error abort ;

: internal-error  cr .in ." You found a bug in gray" gray-error ;

\ contains execution address of a word < token -- > to print a token

variable print-token ' . print-token !

\ Temporary definition for print-token

:noname  ( n -- )
   space dup bl <=
   if .
   else dup 126 >
      if drop space
      else emit
      then
   then
; print-token !

: show-tokens
   dup
   if print-token @ apply-to-members
   else drop ." not yet a set"
   then
;

\ print the intersection of set1 and set2 if it isn't empty

: check-conflict  ( set1 set2 -- )
   2dup disjoint? 0=
   if
\      cr ." Set 1: " over show-tokens
\      cr ." Set 2: " dup  show-tokens
      cr .in ." conflict: "
      clone-set tuck intersection dup show-tokens cr cr
      delete      \ temporary set used for the intersection
   else
      2drop
   then
;

\ ---[ General routines ]------------------------------------------------------

\ Set the generator field in a Syntax-Expr class. Used to point the
\ Syntax-Expr to the concrete Generator class

: set-generator  ( class1 class2 -- )  generator ! ;

\ compute the first-set and maybe-empty of a syntax-expr
\ a bit of memoization is used here

: compute  ( syntax-expr -- first-set maybe-empty )
   new-context
   first-set @ 0=
   if
      this compute()     ( -- first-set maybe-empty )
      maybe-empty !
      first-set !
   then
   first-set @ maybe-empty @
   old-context
;

: get-first  ( syntax-expr -- first-set )
   compute drop
;

: check-cycle  ( syntax-expr -- )  \ just check for left recursion )
   compute 2drop
;

\ add follow-set to the follow set of syntax-expr and its children ) 

: propagate  ( follow-set syntax-expr -- )
   new-context
   dup follow-set @ subset? 0=      \ would everything stay the same
   mark-propagate @ 0= or 
   if                               \ and was propagate here already
      true mark-propagate !        \ NO, do propagate
      follow-set @ clone-set 
      tuck union dup follow-set !
      this propagate()
   else
      drop
   then
   old-context
;

: generate  ( syntax-expr -- )   \ this one gets things done )
   new-context
   this get-generator generate()
   old-context
;

\ pass2 computes all necessary first sets, checks for left recursions
\ and conflicts and generates code for rules

: pass2  ( syntax-expr -- )
   new-context mark-pass2 @ 0=
   if
      true mark-pass2 !
      this check-cycle
      this pass2()
   then
   old-context
;

\ ---[ Main routine ]----------------------------------------------------------

\ use: syntax-expr parser xxx where syntax-expr is for a non-terminal

: parser  ( syntax-expr -- )
   context-stack clear
   new-context
   empty this propagate
   this pass2
   this get-generator begin-def
   this generate
   this get-generator end-def
   old-context
;

\ : parser  ( syntax-expr -- )
\   context-stack clear
\   empty over propagate
\   dup pass2
\   >r r@ new-context
\   r@ get-generator begin-def
\   r@ generate
\   r> get-generator end-def
\   old-context
\ ;

\ ---[ eps - empty syntax-expr ]-----------------------------------------------

Syntax-Expr class
end-class Eps-Syntax-Expr

\ eps1 cannot be set at compile time as max-member will not have been called
\ when "Eps-Syntax-Expr new" crashes. Hence do it first time eps is executed

0 value eps1

:noname  ( class -- syntax-expr )
   eps1 0=
   if
      [ Syntax-Expr :: new ]
      tokset new over (first-set) !    \ Can we use empty here?
      tokset new over (follow-set) !   \ and here?
      true over (maybe-empty) !
      to eps1
   else
      drop
   then
   eps1
; Eps-Syntax-Expr defines new

' internal-error    Eps-Syntax-Expr defines compute()
' drop-args         Eps-Syntax-Expr defines propagate()
' no-syntax-expr-op Eps-Syntax-Expr defines pass2()

\ !! Investigate to see whether the first field etc can ever change in eps1
\ if not move their initialisation into Eps-Syntax-Expr::new 

: eps  ( -- syntax-expr )  \ just adjusts eps1 and returns it
   Eps-Syntax-Expr new
;

\ ---[ Terminals ]-------------------------------------------------------------
\ a terminal is a syntax-expr with extra fields

Syntax-Expr class
   cell var (check&next)   \ contains xt of a word < f -- > that checks
                           \ if f is true and reads the next terminal symbol
   2 cells var (check&next-name) \ Holds the name of a word for the above
end-class Terminal-Syntax-Expr

: check&next  this (check&next) @ ;
: check&next-name this (check&next-name) 2@ ;

\ Original Gray needs an xt for execution of a word. The new version that
\ generates Forth source code needs a name. To maintain compatibility with
\ the original Gray, there are two possible stack inputs to the following.
\ If xt is 0 then there is a string (caddr u) on the stack to provide a
\ reference to the name to be used.  

:noname  ( first-set xt class -- syntax-expr ) \ or
         ( first-set caddr u 0 class -- syntax-expr )
   [ Syntax-Expr :: new ] >r  ( -- first-set xt | first-set caddr u 0 )
   dup if 0 0 rot then        ( -- first-set 0 0 xt | first-set caddr u 0 )
   r@ (check&next) !          ( -- first-set caddr u )
   r@ (check&next-name) 2!    ( -- first-set )
   r@ (first-set) !           ( -- )
   r>                         ( -- syntax-expr )
; Terminal-Syntax-Expr defines new

' internal-error Terminal-Syntax-Expr defines compute()
' drop-args      Terminal-Syntax-Expr defines propagate()

' no-syntax-expr-op Terminal-Syntax-Expr defines pass2()

: terminal  ( first-set xt -- ) \ or ( first-set caddr u 0 -- )
   Terminal-Syntax-Expr new constant
;

\ ---[ Binary syntax-exprs ]---------------------------------------------------

Syntax-Expr class
   cell var (operand1)
   cell var (operand2)
end-class Binary-Syntax-Expr

: operand1  this (operand1) @ ;
: operand2  this (operand2) @ ;

:noname  ( syntax-expr1 syntax-expr2 class -- syntax-expr )
   [ Syntax-Expr :: new ]     ( -- syntax-expr1 syntax-expr2 syntax-expr )
   tuck (operand2) !
   tuck (operand1) !          ( -- syntax-expr )
; Binary-Syntax-Expr defines new

:noname  ( syntax-expr -- )
   drop
   operand1 pass2
   operand2 pass2
; Binary-Syntax-Expr defines pass2()

\ ---[ Concatenations ]--------------------------------------------------------

Binary-Syntax-Expr class
end-class Concat-Syntax-Expr

:noname  ( syntax-expr -- first maybe-empty )
   drop operand1 compute
   dup
   if
      drop
      operand2 compute
      >r clone-set tuck union r>
   then
; Concat-Syntax-Expr defines compute()

:noname  ( follow-set syntax-expr -- )
   drop operand2 compute
   if
      over clone-set tuck union
   then                          ( follow follow1 )
   operand1 propagate
   operand2 propagate
; Concat-Syntax-Expr defines propagate()

\ pass2() is inherited

\ concat is the actual concatenation operator but for safety and readability
\ the parenthesised notation is preferred

: concat  ( syntax-expr1 syntax-expr2 -- syntax-expr )
   Concat-Syntax-Expr new
;
\ ---[ Alternatives ]----------------------------------------------------------

Binary-Syntax-Expr class
end-class Alt-Syntax-Expr

:noname  ( syntax-expr -- first maybe-empty )
   drop
   operand1 compute
   operand2 compute
   rot 2dup and
   if
      cr .in ." Warning: two branches may be empty"
   then
   or >r clone-set tuck union r>
; Alt-Syntax-Expr defines compute()

:noname  ( follow syntax-expr -- )
   drop dup operand1 propagate
   operand2 propagate
; Alt-Syntax-Expr defines propagate()

:noname  ( syntax-expr -- )
   this compute
   if
      follow-set @ check-conflict
   else
      drop
   then
   operand1 get-first operand2 get-first check-conflict
   [ Binary-Syntax-Expr :: pass2() ]
; Alt-Syntax-Expr defines pass2()

\ alt is the actual alternative operator but for safety and readability the
\ parenthesised notation is preferred

: alt  ( syntax-expr1 syntax-expr2 -- syntax-expr )
   Alt-Syntax-Expr new
;

\ ---[ Unary syntax-exprs ]----------------------------------------------------

Syntax-Expr class
   cell var (operand)
end-class Unary-Syntax-Expr

: operand  ( -- syntax-expr )  this (operand) @ ;

:noname  ( syntax-expr1 class -- syntax-expr2 )
   [ Syntax-Expr :: new ]        ( -- syntax-expr1 syntax-expr2 )
   tuck (operand) !              ( -- syntax-expr2 )
; Unary-Syntax-Expr defines new

\ ---[ Options and repetitions ]-----------------------------------------------

:noname ( syntax-expr -- )
   drop
   follow-set @ operand get-first check-conflict
   operand pass2
; Unary-Syntax-Expr defines pass2()

\ ---[ Options ]---------------------------------------------------------------

Unary-Syntax-Expr class
end-class Option-Syntax-Expr

:noname  ( syntax-expr -- set f )
   drop operand compute
   if
      cr .in ." Warning: unnecesary option"
   then
   true
; Option-Syntax-Expr defines compute()

:noname  ( follow syntax-expr -- )
   drop operand propagate
; Option-Syntax-Expr defines propagate()

\ pass2() is inherited
 
: ??  ( syntax-expr1 -- syntax-expr2 )
   Option-Syntax-Expr new
;

\ ---[ Repetitions ]-----------------------------------------------------------

Unary-Syntax-Expr class
end-class Repetition-Syntax-Expr

:noname  ( follow-set syntax-expr -- )
   drop
   operand get-first clone-set tuck union operand propagate
; Repetition-Syntax-Expr defines propagate()

\ ---[ *-repetitions ]---------------------------------------------------------

Repetition-Syntax-Expr class
end-class *Repetition-Syntax-Expr

:noname  ( syntax-expr -- set f )
   drop operand compute
   if
      cr .in ." Warning: *repetition of optional term"
   then
   true
; *Repetition-Syntax-Expr defines compute()

\ propagate() is inherited
\ pass2() is inherited

: **  ( syntax-expr1 -- syntax-expr2 )
   *Repetition-Syntax-Expr new         \ Executes Unary-Syntax-Expr::new
 ;

\ ---[ +-repetitions ]---------------------------------------------------------

Repetition-Syntax-Expr class
end-class +Repetition-Syntax-Expr

:noname  ( syntax-expr -- set f )
   drop operand compute
; +Repetition-Syntax-Expr defines compute()

\ propagate() is inherited
\ pass2() is inherited

: ++  ( syntax-expr1 -- syntax-expr2 )
   +Repetition-Syntax-Expr new         \ Executes Unary-Syntax-Expr::new
;

\ ---[ Actions ]---------------------------------------------------------------

Syntax-Expr class
   Cell var (action)
end-class Action-Syntax-Expr

: action  this (action) @ ;

:noname  ( class -- syntax-expr addr)
   [ Syntax-Expr :: new ]
   empty over (first-set) !
   true over (maybe-empty) !
   dup (action)
; Action-Syntax-Expr defines new

' internal-error    Action-Syntax-Expr defines compute()
' drop-args         Action-Syntax-Expr defines propagate()
' no-syntax-expr-op Action-Syntax-Expr defines pass2()

: newAction  ( class -- syntax-expr i*x )
   new                           ( -- syntax-expr addr )
   over dup get-generator begin-action
;

: {{  ( -- syntax-expr i*x )  \ i*x depends on the generator
   Action-Syntax-Expr newAction
;

: }}  ( syntax-expr i*x -- syntax-expr )
   this get-generator end-action
; immediate

\ For BNF input file where actions are expressed as {foo ...bar}
\ Defined as a separate class to provide access to different action
\ generators, methods are all inherited

Action-Syntax-Expr class
end-class BNF-Action-Syntax-Expr

\ BNF provided for Grace action when { has been recognised

: BNF-action  ( -- syntax-expr i*x )  \ i*x depends on the generator
   BNF-Action-Syntax-Expr newAction
;

\ ---[ Non-terminals ]---------------------------------------------------------

Syntax-Expr class
   cell var (mark-compute)
   cell var (rule-body)       \ in forth left side of rule
   cell var (exec)            \ xt of code for rule
   cell var (deferred)        \ Forward reference needed
   2 cells var (nt-name)      \ (caddr u) of terminal name
end-class nt-Syntax-Expr

: mark-compute  this (mark-compute) ;
: rule-body     this (rule-body)    ;
: exec          this (exec)         ;
: deferred      this (deferred)     ;
: nt-name       this (nt-name)      ;

:noname  ( addr class -- nt-syntax-expr )
   [ Syntax-Expr :: new ]
   false over (mark-compute) !
   tuck (rule-body) !
   0 over (exec) !
   0 over (deferred) !
; nt-Syntax-Expr defines new

\ get the body of the rule for the nt in "this"

: get-body  ( -- syntax-expr )
   rule-body @
   if
      rule-body @
   else
      cr .in ." No rule for nonterminal" gray-error
   then
;

:noname  ( syntax-expr -- set f )
   drop mark-compute @
   if
      cr .in ." Left recursion" gray-error
   else
      true mark-compute !
      get-body compute
   then
; nt-Syntax-Expr defines compute()

:noname ( follow-set syntax-expr -- )
   drop get-body propagate 
; nt-Syntax-Expr defines propagate()

\ In addition to the usual duties, this pass2() also has to code-nt

:noname  ( syntax-expr -- )
   >r get-body pass2
   r> get-generator code-nt
; nt-Syntax-Expr defines pass2()

: make-nt  ( syntax-expr -- nt-syntax-expr )
   nt-Syntax-Expr new
;

\ <- saves the non-terminal name in dataspace as <name> for
\ a Forth source code generation call to the non-terminal

: <-           \ use: syntax-expr <- xxx
               \ xxx: -- syntax-expr
   >in @ >r make-nt dup constant ( -- syntax-expr )
   r> >in ! parse-name           ( -- syntax-expr caddr u )
   here over 2>r
   [char] < c,
   here swap move r@ allot    ( -- syntax-expr )
   [char] > c,
   2r> 2 chars + rot (nt-name) 2!   ( -- )
;

: nonterminal  \ use: nonterminal xxx
   0 <-        \ forward declaration
;

\ ---[ Rules ]-----------------------------------------------------------------

: rule  ( syntax-expr nt -- )    \ Makes a rule )
   new-context
   rule-body @
   if
      .in ." multiple rules for nonterminal" gray-error
   then
   rule-body !
   old-context 
;

\ ---[ Syntactic sugar ]-------------------------------------------------------

\ Example: 0 5 6 7 ' + reduce  =  5 6 7 + +  =  18 )
\ The original version of reduce generated a linear sequence of IF statements,
\ the 2 alternatives generate a tree of IFs

0 [if]   \ Original Gray version
	: reduce  ( 0 x1 ... [x2 x3 -- x4] -- x )
	   >r dup 0=
	   if ." no operand" abort then
	   begin
	      over
	   while
	      r@ execute
	   repeat            ( -- 0 x )
	   swap drop r> drop
	;
[then]

1 [if]  \ The best comp.lang.forth version
   : reduce-1level ( 0 x1 ... xn xt -- 0 y1 ... ym xt )  \ m=(n+1)/2 
      over
      if 
         2 pick
         if 
            dup >r execute r> swap >r recurse r> swap 
         then 
      then
;
 
   : reduce ( 0 x1 ... xn xt -- x ) 
      over 0= abort" no operand" 
      begin 
         2 pick
      while 
         reduce-1level 
      repeat 
      drop nip
   ; 
[then]

0 [if]   \ GWJ version, using ROLL - slower
: count-operands ( 0 x1 ... xu -- 0 x1 ... xu u+1 ) 
    1 begin 
        dup pick
      while 
            1+
      repeat 
;

: reduce  ( 0 x1 ... xn xt -- x )
   over 0= abort" no operand" 
   >r count-operands
   begin
      dup roll dup              
      if
         over roll dup
         if r@ execute swap 1- else rot then
      else
         swap
      then
      dup 2 =
   until
   r> 2drop ?dup if nip then
;

[then]

7 constant concatenation-id

: (-  ( -- n 0 )
   concatenation-id 0
;

: -)  ( n 0 syntax-expr1 syntax-expr2 .. -- syntax-expr )
   ['] concat reduce
   swap concatenation-id ?pairs
;

8 constant alternative-id

: (|  ( -- n 0 )
   alternative-id 0
;

: |)  ( n 0 syntax-expr1 syntax-expr2 .. -- syntax-expr )
   ['] alt reduce
   swap alternative-id ?pairs
;

: ((  (| (- ;
: ))  -) |) ;
: ||  -) (- ;

\ ---[ Testing ]---------------------------------------------------------------
development [if]
: .sets  ( syntax-expr -- )
   new-context    ( -- )
   cr ." First set: " first-set @ show-tokens
   cr ." Follower set: " follow-set @ show-tokens
   cr ." Maybe empty: " maybe-empty @
      if ." can " else ." cannot " then ." be empty"
   cr cr
   old-context
;

: .common
   cr ." First set: " first-set @ dup . show-tokens
   cr ." Follow set: " follow-set @ dup . show-tokens
   cr ." Maybe empty: " maybe-empty @
   if ." can " else ." cannot " then ." be empty"
;

: .term  ( syntax-expr -- )
   new-context                   ( -- )
   .common
   cr ." Check&next: " check&next .
   cr cr
   old-context
;

: .bin  ( syntax-expr -- )
   new-context
   .common
   cr ." Operand 1: " operand1 .
   cr ." Operand 2: " operand2 .
   cr cr
   old-context
;

: .unary  ( syntax-expr -- )
   new-context
   .common
   cr ." Operand: " operand .
   cr cr
   old-context
;

: .action  ( syntax-expr -- )
   new-context
   .common
   cr ." Action: " action .
   cr cr 
   old-context
;

: .nonterm  ( syntax-expr -- )
   new-context
   .common
   cr ." Mark compute: " mark-compute @ if ." True" else ." False" then
   cr ." Rule: " rule-body @ .
   cr ." Exec: " exec @ .
   cr cr
   old-context
;
[then]
\ -----------------------------------------------------------------------------

.( graycore.fth loaded. ) .s
