\ The concrete generator classes to emulate the original version of Gray
\ i.e. to compile immediately executable code

\ This software was taken from gray.fs with this copyright notice

\ Copyright 1990, 1991, 1994, 2003 Martin Anton Ertl

\     This program is free software; you can redistribute it and/or modify
\     it under the terms of the GNU General Public License as published by
\     the Free Software Foundation; either version 2 of the License, or
\     (at your option) any later version.

\     This program is distributed in the hope that it will be useful,
\     but WITHOUT ANY WARRANTY; without even the implied warranty of
\     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
\     GNU General Public License for more details.

\     You should have received a copy of the GNU General Public License
\     along with this program; if not, write to the Free Software
\     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

\ -----------------------------------------------------------------------------
\ The software has been extended and the extensions are 
\ copyright (C) Gerry Jackson 2008
\ -----------------------------------------------------------------------------
.( Loading graygen.fth ... ) cr

\ ---[ Tests ]-----------------------------------------------------------------

\ The user must set test-vector to the execution token of a user-defined word
\ ( set -- f ) that returns true if the token of the current symbol is in set

variable test-vector
:noname
   -1 abort" test-vector has not been initialised"
; test-vector !

: compile-test  ( set -- )
   postpone literal
   test-vector @ compile,
;

\ ---[ Common base class definitions ]-----------------------------------------

:noname  ( class -- [colon-sys] )    \ Some forths may leave colon-sys
 \ : should not be immediate, but just for the case it is, we use [compile]
 \ if your system does not have [compile], remove it
   drop [compile] : 
; dup GeneratorBase defines begin-def
  dup ActionGeneratorBase defines begin-def
      Pass2GeneratorBase defines begin-def

:noname  ( [colon-sys] class -- )
   drop postpone ;
; dup GeneratorBase defines end-def
  dup ActionGeneratorBase defines end-def
      Pass2GeneratorBase defines end-def

\ ---[ eps generator ]-----------------------------------------------

GeneratorBase class
end-class Gray-Eps-Generator

' no-syntax-expr-op Gray-Eps-Generator defines generate()

Gray-Eps-Generator Eps-Syntax-Expr set-generator
 
\ ---[ Terminal generator ]-----------------------------------------------

GeneratorBase class
end-class Gray-Terminal-Generator

:noname  ( gen-class -- )
   drop this get-first compile-test
   check&next dup 0=
   if -1 abort" Error: check&next word unspecified for terminal" then
   compile,
; Gray-Terminal-Generator defines generate()

Gray-Terminal-Generator Terminal-Syntax-Expr set-generator

\ ---[ Concatenation generator ]-----------------------------------------------

GeneratorBase class
end-class Gray-Concat-Generator

:noname  ( gen-class -- )
   drop
   operand1 generate
   operand2 generate
; Gray-Concat-Generator defines generate()

Gray-Concat-Generator Concat-Syntax-Expr set-generator

\ ---[ Alternative generator ]-----------------------------------------------

GeneratorBase class
end-class Gray-Alt-Generator

: generate-alternative1  ( -- )
   operand1 get-first compile-test
   postpone if
   operand1 generate
   postpone else
   operand2 generate
   postpone then
;

: generate-alternative2  ( -- )
   operand1 get-first compile-test postpone 0=
   operand2 get-first compile-test postpone and
   postpone if
   operand2 generate
   postpone else
   operand1 generate
   postpone then
;

:noname  ( gen-class -- )
   drop operand1 compute
   if
      generate-alternative2
   else
      generate-alternative1
   then
   drop
; Gray-Alt-Generator defines generate()

Gray-Alt-Generator Alt-Syntax-Expr set-generator

\ ---[ Option generator ]-----------------------------------------------

GeneratorBase class
end-class Gray-Option-Generator

:noname  ( gen-class -- )
   drop operand get-first compile-test
   postpone if
   operand generate
   postpone then
; Gray-Option-Generator defines generate()

Gray-Option-Generator Option-Syntax-Expr set-generator

\ ---[ *Repetition generator ]-----------------------------------------------

GeneratorBase class
end-class Gray-*Repetition-Generator

:noname  ( gen-class -- )
   drop postpone begin
   operand get-first compile-test
   postpone while
   operand generate
   postpone repeat
; Gray-*Repetition-Generator defines generate()

Gray-*Repetition-Generator *Repetition-Syntax-Expr set-generator

\ ---[ +Repetition generator ]-----------------------------------------------

GeneratorBase class
end-class Gray-+Repetition-Generator

:noname  ( gen-class -- )
   drop
   postpone begin
   operand generate
   operand get-first compile-test
   postpone 0= postpone until
; Gray-+Repetition-Generator defines generate()

Gray-+Repetition-Generator +Repetition-Syntax-Expr set-generator

\ ---[ Action generators ]------------------------------------------------------

ActionGeneratorBase class
end-class Gray-Action-Generator

:noname  ( syntax-expr gen-class2 -- )
   drop action compile,
; Gray-Action-Generator defines generate()

:noname  ( syntax-expr gen-class -- xt [colon-sys] )
   drop new-context :noname             \ old-context is in Action::end-action
; Gray-Action-Generator defines begin-action

:noname  ( syntax-expr ad xt [colon-sys] class -- syntax-expr)
   drop postpone ; swap ! old-context
; Gray-Action-Generator defines end-action

Gray-Action-Generator Action-Syntax-Expr set-generator

\ For conventional BNF input. This starts a :noname definition, scans text input
\ and evaluates (compiles) the action until } has been found. The } symbol is
\ consumed and end-action should never be called. Note that BL WORD can't be
\ used for scanning as there may not be a space before the terminating }

Gray-Action-Generator class
end-class Gray-BNF-Action-Generator

:noname  ( syntax-expr ad syntax-expr gen-class -- syntax-expr )
   drop new-context :noname            ( -- syntax-expr ad xt )
   begin
      source >in @ /string             ( -- syntax-expr ad xt caddr1 u1 )
      2dup s" }" search                ( -- syntax-expr ad xt caddr1 u1 caddr2 u2 f )
      >r swap source drop - >in ! r@   ( -- syntax-expr ad xt caddr1 u1 u2 f )
      if - else drop then              ( -- syntax-expr ad xt caddr1 u3 )
      evaluate r> 0=                   ( -- syntax-expr ad xt f2 )
   while
      refill 0=
   until then
   postpone ; swap ! old-context       ( -- syntax-expr )
; Gray-BNF-Action-Generator defines begin-action

' internal-error Gray-BNF-Action-Generator defines end-action

Gray-BNF-Action-Generator BNF-Action-Syntax-Expr set-generator



\ ---[ Non-terminal generator ]------------------------------------------------

Pass2GeneratorBase class
end-class Gray-nt-Generator

:noname  ( gen-class -- )
   drop exec postpone literal
   postpone @
   postpone execute
; Gray-nt-Generator defines generate()

:noname  ( class -- )    \ generates the code for a rule
   drop :noname 
   get-body generate
   postpone ;
   exec !
; Gray-nt-Generator defines code-nt

Gray-nt-Generator nt-Syntax-Expr set-generator

\ -----------------------------------------------------------------------------

.( graygen.fth loaded ) .s
