\ Files.fth    31/12/2007 - under development

\ Provides higher level words than the ANS file set of words to write to and
\ read from files

\ -----------------------------------------------------------------------------
\ Copyright (C) Gerry Jackson 2008

\ This software is free; you can redistribute it and/or modify it in
\ any way provided you acknowledge the original source and copyright

\ This program is distributed in the hope that it will be useful,
\ but WITHOUT ANY WARRANTY; without even the implied warranty of
\ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
\ -----------------------------------------------------------------------------

.( files.fth loading ...) cr

\ -----------------------------------------------------------------------------
\ To enable multiple files to be opened and used a File class will be defined
\ whose objects will hold relevant file data. Initially, at least, a current
\ file variable will be used to avoid the user having to continually specify
\ the file to be used. Later it may be desirable to provide a version of the
\ read/write words where the user has to specify the File object to be used.

\ Requires xmini_oof to have been loaded

object class
   1 cells var fid
end-class File

: get-fid  ( file -- fid )  fid @ ;

\ -----------------------------------------------------------------------------
\ Current file handling

0 value currentFile

: setCurrentFile  ( file -- ) to currentFile ;

: getCurrentFile  ( -- file ) postpone currentFile ; immediate

: currentFile  ( -- file )
   getCurrentFile ?dup 0=
   abort" Current file is undefined"
;
   

\ -----------------------------------------------------------------------------
\ Create, open and close files

: newFile  ( fid -- file )
   File new tuck fid !
;

: createFile  ( caddr u -- file )
   w/o create-file               ( -- fid ior )
   abort" Unable to create a file"
   newFile                       ( -- file )
;

\ open a file for read and/or write

: openFile  ( caddr u -- file )
   r/w open-file                 ( -- fid ior )
   abort" Unable to open a read/write file"
   newFile                       ( -- file )
;

\ Close the current file

: closeFile    ( -- )
   currentFile dup delete        ( -- file )
   fid @ close-file              ( -- ior )
   abort" Unable to close output file"
;

\ -----------------------------------------------------------------------------
\ Write text and numbers to the current file

: writeText  ( caddr u -- )
   currentFile fid @ write-file
   abort" Writing text to file failed"
;

: writeLine  ( caddr u -- )
   currentFile fid @ write-line
   abort" Writing a line of text to file failed"
;

: writeNewLine  ( -- ) 0 0 writeLine ;

: writeSpace  ( -- ) s"  " writeText ;

: writeSpaces  ( n -- ) 0 ?do writeSpace loop ;

: writeNumber.R   ( n u -- )
   over 2>r abs 0 <# #s r> sign #>
   r> over - 0 max writeSpaces writeText
;

: writeNumberU.R  ( u1 u2 -- )
   >r 0 <# #s #>
   r> over - 0 max writeSpaces writeText
;

: writeNumber  ( n -- ) 0 writeNumber.R ;

\ -----------------------------------------------------------------------------

.( files.fth loaded. ) .s
