\ List library file

\ Still in development

\ -----------------------------------------------------------------------------
\ Copyright (C) Gerry Jackson 2008

\ This software is free; you can redistribute it and/or modify it in
\ any way provided you acknowledge the original source and copyright

\ This program is distributed in the hope that it will be useful,
\ but WITHOUT ANY WARRANTY; without even the implied warranty of
\ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
\ -----------------------------------------------------------------------------

.( List.fth loading ...) cr

\ -----------------------------------------------------------------------------
\ List item class

object class
   1Cell var next-item
   method compare-item
end-class ListItem

\ compare-item must be defined by the caller for list_find to work. This is
\ because only the user knows the type of the contents of the list item. 
\ compare-item must compare list item values and return 0 for a match 
\ otherwise <>0. The default is to report an error

:noname  ( item list -- f )
   abort" compare-item has not been defined"
; ListItem defines compare-item

\ -----------------------------------------------------------------------------
\ List class - for a list of list items

object class
   1Cell var first-item
   1Cell var last-item
   1Cell var item-count   \ Number of items in the list
   1Cell var prev-item    \ Previous item, saved by find-item
   1Cell var item-class   \ The class of list items
   method get-head
   method get-tail
   method list-insert     \ Insert at head of list
   method list-append     \ Append to end of list
   method is-empty
   method list-remove     \ Remove list item
   method list-find       \ Search for list item
   method get-item-position
   method list-length     \ Number of items in the list
   method list-display
end-class List

\ An empty list contains a dummy item, pointed to by first-item and last-item

:noname  ( class -- list )
   [ object :: new ]                ( -- list )
   >r ListItem new                  ( -- item ) \ Dummy item
   0 over next-item !
   dup r@ first-item !
   dup r@ last-item !
   r@ prev-item !                   ( -- )
   0 r@ item-count !
   r>                               ( -- list )
; List defines new

\ get-head returns the first list item, leaving it in the list

:noname  ( list -- item )  \ item = 0 for empty list
   first-item @ next-item @         ( -- item )
; List defines get-head

:noname  ( list -- item )  \ item = 0 for empty list
   dup item-count @
   if last-item @ else 0= then      ( -- item | 0 )
; List defines get-tail

\ Insert item at the head of the list

:noname  ( item list -- )
   dup >r first-item @ dup >r       ( -- item dum ) ( R: -- list dum )
   next-item @ over next-item !     ( -- item )
   dup r> next-item !               ( -- item )  ( R: -- list )
   r@ item-count @
   if drop else r@ last-item ! then ( -- )
   1 r> item-count +!
; List defines list-insert

:noname  ( item list -- )
   1 over item-count +!
   last-item >r dup r@ @            ( -- item item item2 ) ( R: -- ad1 )
   next-item !                      ( -- item )
   0 over next-item !
   r> !                             ( -- )
; List defines list-append

:noname  ( list -- f )
   item-count @ 0=     \ Alternatively test first-item = last-item
; List defines is-empty

\ Remove item from the list, but do not delete it.
\ Error if item is not in the list, includes an empty list

:noname  ( item list -- flag )
   2dup >r >r                       ( R: -- list item ) 
   tuck prev-item @ next-item @ <>  ( -- list f )
   if
      first-item @ dup next-item @  ( -- prev curr ) \ prev is dummy
      begin
         ?dup
      while
         dup r@ <>
      while
         nip dup next-item @        ( -- prev' curr' )
      repeat
          \ Found it
      else
         r> 2drop r@ first-item @   ( -- dum )
         r> prev-item !             ( -- )
         -1 abort" List item not found"   \ For now
      then
   else
\      cr ." Using prev" .s
      prev-item @ r@                ( -- prev item )      
   then
   next-item @                      ( -- prev next ) \ Found item
   swap next-item !                 ( -- )
   r> drop
   r@ first-item @ r@ prev-item !
   -1 r> item-count +! 
; List defines list-remove

:noname     ( list -- item | 0 )
   dup >r
   first-item @ dup next-item @     ( -- prev curr )
   begin
      dup
   while
      tuck 2>r                      ( -- curr )
      compare-item                  ( -- f )
   while
      2r> nip                       ( -- curr )
      dup next-item @               ( -- prev' curr' )
   repeat
      2r> swap                      ( -- curr' prev')
   else           \ Not found
      nip r@ first-item @           ( -- 0 dum ) 
   then
   r> prev-item !                   ( -- item | 0 )
; List defines list-find

:noname  ( item list -- pos )       \ returns -1 if not found
   0 >r get-head                    ( -- item item2 ) ( R: -- pos )
   begin
      dup
   while
      2dup = if 2drop r> exit then
      r> 1+ >r
      next-item @
   repeat
   r> drop 2drop -1
; List defines get-item-position

:noname  ( list -- u )  item-count @ ; List defines list-length

\ -----------------------------------------------------------------------------

.( List.fth loaded. ) .s

