\ miniactions_gray.fth for use with the mini language example
\ It handles the mode with the grammar in Gray format and compilation of
\ the parser in memory i.e. Gray emulation mode

\ 26/9/08 Modified to use LexScanner 2.0

.( Loading miniactions_gray.fth ... ) cr

\ ---[ Optional screen display ]------------------------------------------------
\ Depends on user settable flag verbose

: ?cr    ( -- )        verbose if cr then ;
: ?space ( -- )        verbose if space then ;
: ?type ( caddr u -- ) verbose if type else 2drop then ;

\ ---[ Interface to the scanner ]-----------------------------------------------

BaseDefaultData #states CheckNextData maxCheck init-scanner value mini-scanner

: getNextToken  ( -- caddr u tok )
   begin
      mini-scanner next-token dup white_space =
   while
      ?space drop 2drop
   repeat
;

\ ---[ Mini program source file handling  ]-------------------------------------

0 value mini-source

: open-mini-source  ( caddr u -- )
   r/o open-file throw 
   file-source dup set-source
   to mini-source
;

: close-mini-source  ( -- )
   mini-scanner close-scanner
   source-id close-file throw
   set-ANS-source
   mini-source close-source 0 to mini-source
;

\ ---[ Shared variables ]-------------------------------------------------------

variable sym
2variable symname
variable identp
variable identlen

\ ---[ Interface to Gray ]------------------------------------------------------

tokenval @ 1- max-member      \ The token set size

: testsym? ( set -- f )
   sym @ member?
;

' testsym? test-vector !

: >ident-string   ( caddr u -- ) \ Factor of ?nextsym
   here identp ! dup identlen !
   here swap move
;

: ?nextsym  ( f -- )    \ Used for check&next in Gray terminals
   0= if close-mini-source -1 abort" Syntax error" then
   begin
      getNextToken dup 0<     ( -- caddr u tok f )
   while
      drop 2drop refill 0=    ( -- f2 )
   until
      source eof_token
   then
   dup ident = if >r 2dup >ident-string r> then
   sym ! 2dup ?type symname 2!
;

\ The call to terminal is after definition of the embedded actions in the
\ definition of token

\ ---[ Actions embedded in the grammar and helpers ]----------------------------
 
: ident-string ( -- addr u )
   identp @ identlen @
;

: append-string ( addr1 u1 addr2 u2 -- addr1 u3 )
   3 roll >r            ( u1 a2 u2 )
   rot 2dup +           ( a2 u2 u1 u1+u2 )
   r> swap >r           ( a2 u2 u1 a1 )
   dup >r
   + swap move
   r> r>
;

create evalbuf 128 allot      \ Using PAD is unsafe

\ Defines colon-def whose name is give by (addr u)

: :name ( addr u -- colon-sys )
   evalbuf 0 s" : " append-string 2swap append-string evaluate
;

wordlist constant variable-wordlist \ for mini variables

\ Defines variable whose name is given by (addr u)

: variable-name ( addr u -- )
   2dup variable-wordlist search-wordlist abort" variable already defined"
   get-current >r variable-wordlist set-current
   evalbuf 0 s" variable " append-string 2swap append-string evaluate
   r> set-current
;

\ Get the execution token of the var whose name is given by (addr u)

: getvar ( addr u -- xt )
   variable-wordlist search-wordlist 0= abort" variable undefined"
;

: getNumber    ( -- n )
   0 0 symname 2@ >number 2drop drop   ( -- n )
;

\ ---[ Redefinition of token ]--------------------------------------------------
\ token has to be redefined when the tokens file is read for a second time
\ during the loading of the Gray system. This new definition will create
\ Gray terminals with the same name as the tokens. It gets the token value
\ for each terminal from the constant with the same name. As these constants
\ may be used in parsing actions above, this redefinition must take place
\ after the actions have been defined. The words singleton and terminal are
\ part of the Gray interface. 

: token
   >in @ ' execute     
                ( -- n1 n2 )
   swap >in !                       ( -- n2 )
   singleton ['] ?nextsym terminal  ( -- )
;

\ ---[ To compile a mini program ]----------------------------------------------

0 value parser-name  \ This should be set by the Gray source file

: mini ( caddr u -- )
   ?cr open-mini-source refill
   if
      true ?nextsym parser-name execute
   else
      cr ." Empty BNF file"
   then
   close-mini-source
;

\ Following compilation of the mini program, run it with a call to the name of
\ the mini program with suitable arguments on the stack

\ ------------------------------------------------------------------------------

.( miniactions_gray.fth loaded. ) .s
