\ minirun.fth for use with the mini language example
\ It provides: the code to run the parser generated by ForthGen


\ 26/9/08 Modified to use lexscanner2_0

.( Loading minirun.fth ... ) cr

\ ---[ Optional screen display ]------------------------------------------------
\ Depends on user settable flag verbose

: ?cr    ( -- )        verbose if cr then ;
: ?space ( -- )        verbose if space then ;
: ?type ( caddr u -- ) verbose if type else 2drop then ;

\ ---[ Interface to the scanner ]-----------------------------------------------

BaseDefaultData #states CheckNextData maxCheck init-scanner value mini-scanner

: getNextToken ( -- caddr u tok )
   begin
      mini-scanner next-token dup white_space =
   while
      ?space drop 2drop
   repeat
;

\ ---[ Mini program source file handling  ]-------------------------------------

0 value mini-source

: open-mini-source  ( caddr u -- )
   r/o open-file throw 
   file-source dup set-source
   to mini-source
;

: close-mini-source  ( -- )
   mini-scanner close-scanner
   source-id close-file throw
   set-ANS-source
   mini-source close-source 0 to mini-source
;

\ ---[ Shared variables ]-------------------------------------------------------

variable sym
2variable symname
variable identp
variable identlen

\ ---[ Interface to the mini parser ]-------------------------------------------

\ Values to be set by the Mini parser

0 value first-set
0 value bytes/set
0 value bits/cell
0 value parser-name

\ Test if sym has its bit number set in the bit vector
\ Written for sets <= 1 cell in size in this example

: testsym? ( set-index -- f )
   bytes/set * first-set + @     ( -- vec )
   1 sym @ lshift and
;

\ An alternative definition for multi-cell sets (>32 members)
\ is given here

0 [if]
: testsym?  ( set-index -- f )
   bytes/set * first-set + >r          ( -- )     ( R:  -- ad )
   sym @ bits/cell /mod cells r> + @   ( -- bit vec )
   1 rot lshift and                    ( -- f )
;
[then]

: test-token  ( n -- f )  sym @ = ;

: >ident-string   ( caddr u -- )
   here identp ! dup identlen !
   here swap move
;

: ?nextsym  ( f -- )    \ Used for check&next in Gray terminals
   0= if close-mini-source -1 abort" Syntax error" then
   begin
      getNextToken dup 0<     ( -- caddr u tok f )
   while
      drop 2drop refill 0=    ( -- f2 )
   until
      source eof_token
   then
   dup ident = if >r 2dup >ident-string r> then
   sym ! 2dup ?type symname 2!
;

\ ---[ Actions embedded in the grammar and helpers ]----------------------------
 
: ident-string ( -- addr u )
   identp @ identlen @
;

: append-string ( addr1 u1 addr2 u2 -- addr1 u3 )
   3 roll >r            ( u1 a2 u2 )
   rot 2dup +           ( a2 u2 u1 u1+u2 )
   r> swap >r           ( a2 u2 u1 a1 )
   dup >r
   + swap move
   r> r>
;

create evalbuf 128 allot      \ Using PAD is unsafe

\ Defines colon-def whose name is give by (addr u)

: :name ( addr u -- colon-sys )
   evalbuf 0 s" : " append-string 2swap append-string evaluate
;

wordlist constant variable-wordlist ( for mini variables )

\ Defines variable whose name is given by (addr u)

: variable-name ( addr u -- )
   2dup variable-wordlist search-wordlist abort" variable already defined"
   get-current >r variable-wordlist set-current
   evalbuf 0 s" variable " append-string 2swap append-string evaluate
   r> set-current
;

\ Get the execution token of the var whose name is given by (addr u)

: getvar ( addr u -- xt )
   variable-wordlist search-wordlist 0= abort" variable undefined"
;

: getNumber    ( -- n )
   0 0 symname 2@ >number 2drop drop   ( -- n )
;

\ ---[ To compile a mini program ]----------------------------------------------

: mini ( caddr u -- )
   ?cr open-mini-source refill
   if
      true ?nextsym parser-name execute
   else
      cr ." Empty Mini source code file"
   then
   close-mini-source
;

\ -----------------------------------------------------------------------------

.( minirun.fth loaded. ) .s
